;; $ID$
;; -*- mode: common-lisp; package: cl-user -*-
;; 
;; Copyright (C) 2004 Korea Lisp User Group
;;
;; WARNING: This code is mainly for educational purpose.
;; That means it may contains intended errors or bugs - use with your
;; own risk.
;;                                    
;; This code is free software; you can redistribute it and/or
;; modify it under the terms of the version 2.1 of
;; the GNU Lesser General Public License.
;; 
;; This code is distributed in the hope that it will be useful,
;; but without any warranty; without even the implied warranty of
;; merchantability or fitness for a particular purpose.  See the GNU
;; Lesser General Public License for more details.
;;
;; Version 2.1 of the GNU Lesser General Public License is in the file
;; license-lgpl.txt that was distributed with this file.
;; If it is not present, you can access it from
;; http://www.gnu.org/copyleft/lesser.txt (until superseded by a newer
;; version) or write to the Free Software Foundation, Inc., 59 Temple Place,
;; Suite 330, Boston, MA  02111-1307  USA
;;


;;;
;;; Utilities
;;;
(defun mapa-b (fn a b &optional (step 1))
  "Collect result of applying function FN from A to B by STEP"
  (loop for i from a to b by step
	collect (funcall fn i)))

(declaim (inline multiple-cons))	; This function can be inline function
(defun multiple-cons (a list-of-list)
  "Cons A to each list of LIST-OF-LIST"
  (mapcar #'(lambda (lst)
	      (cons a lst)) list-of-list))
(declaim (notinline multiple-cons))

(defmacro bind-when ((var pred) &body body)
  "Bind result of PRED to variable VAR, then do BODY in the context of WHEN VAR=TRUE"
  `(let ((,var ,pred))
    (when ,var
      ,@body)))

(defmacro bind-if ((var pred) then else)
  "Bind result of PRED to variable VAR, then do BODY in the context of IF VAR=TRUE THEN ELSE"
  `(let ((,var ,pred))
    (if ,var
	,then
	,else)))
;;;
;;; End of Utilities
;;;

;;;
;;; Classes
;;;

(defclass history-element ()
  ((guess :initarg :guess :accessor guess)
   (ball-count :initform nil :accessor ball-count)))

(defun make-history-element (guess)
  (make-instance 'history-element :guess guess))

(defclass guess-history ()
  ((elements :initform nil :accessor guess-history-elements)))

(defmethod guess-history-last-element ((history guess-history))
  (first (guess-history-elements history)))

(defmethod guess-history-add-element ((history guess-history) (element history-element))
  (push element (guess-history-elements history)))

;; Can we set certain limitation on iterator? (such that it only takes two arguments, ...)
(defmethod guess-history-iterate-element ((history guess-history) iterator)
  (dolist (element (guess-history-elements history))
    (funcall iterator (guess element) (ball-count element))))

;; Base class
(defclass baseball-algorithm ()
  ((guess-history :initform (make-instance 'guess-history) :accessor guess-history)))

;;
;; All subclasses of baseball-algorithm must provide following methods
;;
(defgeneric make-guessing-algorithm (eql-algorithm number-of-digits)
  (:documentation "EQL specializer based constructor. An algorithm must have this method."))

(defgeneric make-random-guess (baseball-algorithm)
  (:documentation "Return a random guess from candidates."))

(defgeneric improve-guess (baseball-algorithm strikes balls)
  (:documentation "Improve current algorithm with given feed back from user - STRIKES and BALLS"))

;;
;; simple-duplicate-permutation-algorithm
;;
(defclass simple-duplicate-permutation-algorithm (baseball-algorithm)
  ((candidates :initarg :candidates :accessor candidates)))    

#||
Writing make-duplicate-permutations-for-baseball-game:

;;; First try
(defun make-duplicate-permutations-for-baseball-game (n)
  (let ((1st-digits (mapa-b #'identity 1 9))
	(nth-digits (mapa-b #'identity 0 9)))
    (labels ((recurse (n result)
	       (if (= n 2)
		   (loop for n in 1st-digits
			 nconc (mapcar #'(lambda (lst)
					   (cons n lst)) result))
		   (recurse (1- n) (loop for n in nth-digits
					 nconc (mapcar #'(lambda (lst)
							   (cons n lst)) result))))))
      (if (= n 1)
	  (mapcar #'list nth-digits)
	  (recurse n (mapcar #'list nth-digits))))))

;;; Second try
(defun make-duplicate-permutations-for-baseball-game (n)
  (let ((1st-digits (mapa-b #'identity 1 9))
	(nth-digits (mapa-b #'identity 0 9)))
    (labels ((recurse (n result)
	       (if (= n 2)
		   (mapcan (lambda (n)
			     (mapcar #'(lambda (lst)
					 (cons n lst)) result))
			   1st-digits)
		   (recurse (1- n) (mapcan (lambda (n)
					     (mapcar #'(lambda (lst)
							 (cons n lst)) result))
					   nth-digits) ))))
      (if (= n 1)
	  (mapcar #'list nth-digits)
	  (recurse n (mapcar #'list nth-digits))))))
||#

(declaim (inline multiple-cons)) ; Make multiple-cons inline function
;;; Third try
;;; Replacing xcombinations
(defun make-duplicate-permutations-for-baseball-game (n)
  "Return duplucated permutation from (0 1 2 3 4 5 6 7 8 9). Lists start with 0 is not included in the result."
  (let ((1st-digits (mapa-b #'identity 1 9))
	(nth-digits (mapa-b #'identity 0 9)))
    (labels ((recurse (n result)
	       (if (= n 2)
		   (mapcan (lambda (n)
			     (multiple-cons n result))
			   1st-digits)
		   (recurse (1- n) (mapcan (lambda (n)
					     (multiple-cons n result))
					   nth-digits)))))
      (if (= n 1)
	  (mapcar #'list nth-digits)
	  (recurse n (mapcar #'list nth-digits))))))
(declaim (notinline multiple-cons))	; Not inline from now on


(defmethod make-guessing-algorithm ((algorithm (eql 'simple-duplicate-permutation-algorithm)) (n integer))
  (make-instance 'simple-duplicate-permutation-algorithm
		 :candidates (make-duplicate-permutations-for-baseball-game n)))

#||
First try
(defmethod make-random-guess ((algorithm simple-duplicate-permutation-algorithm))
  (let ((candidate (pop (candidates algorithm))))
    (when candidate
      (push (make-history-element candidate) (guess-history algorithm))
      candidate)))
||#

;;; Replace make-guessing-algorithm function
(defmethod make-random-guess ((algorithm simple-duplicate-permutation-algorithm))
  (bind-when (candidates (candidates algorithm))
    (let* ((random-position (random (length candidates)))
	   (guess (nth random-position candidates)))
      ;; destructively modify candidates
;;      (setf (nthcdr random-position candidates)
;;	    (nthcdr (1+ random-position) candidates))
      (setf candidates
	    (append (subseq candidates 0 random-position)
		    (subseq candidates (1+ random-position))))
      ;; update guess-history
      (guess-history-add-element (guess-history algorithm) (make-history-element guess))
      
      guess)))

(defun compute-ball-count (last-guess candidate)
  "Return CANDIDATE's similarity to LAST-GUESS as number of strikes and balls. Last-guess and candidate is a list of digit."
  (flet ((number-of-balls (non-strikes-in-guess non-strikes-in-candidate)
	   (loop for guess in non-strikes-in-guess
		 when (find guess non-strikes-in-candidate :test #'=)
		 do (setf non-strikes-in-candidate
			  (delete guess non-strikes-in-candidate :test #'=))
		 and
		 count guess)))
    (labels ((recurse (guess-list candidate-list
		       non-strikes-in-guess non-strikes-in-candidate
		       num-strikes)
	       (if (null guess-list)
		   (values num-strikes (number-of-balls non-strikes-in-guess
							non-strikes-in-candidate))
		   (let ((guess-digit (car guess-list))
			 (candidate-digit (car candidate-list)))
		     (if (= guess-digit candidate-digit)
			 (recurse (cdr guess-list)
				  (cdr candidate-list)
				  non-strikes-in-guess
				  non-strikes-in-candidate
				  (1+ num-strikes))
			 (recurse (cdr guess-list)
				  (cdr candidate-list)
				  (cons guess-digit non-strikes-in-guess)
				  (cons candidate-digit non-strikes-in-candidate)
				  num-strikes))))))
      (recurse last-guess candidate () () 0))))

;; Renamed from feedback
(defmethod improve-guess ((algorithm simple-duplicate-permutation-algorithm) strikes balls)
  ;; Keep ball count history
  (setf (ball-count (guess-history-last-element (guess-history algorithm)))
	(list strikes balls))
  ;; Update candidates
  (let ((last-guess (guess (guess-history-last-element (guess-history algorithm)))))
    (setf (candidates algorithm)
	  (remove-if-not (lambda (candidate)
			   (multiple-value-bind (computed-strikes computed-balls)
			       (compute-ball-count last-guess candidate)
			     (and (= strikes computed-strikes) (= balls computed-balls))))
			 (candidates algorithm)))))

#||
(defun get-printable-history (algorithm)
  (labels ((recurse (history-elements result-guesses strikes balls)
	     (if (null history-elements)
		 (values result-guesses strikes balls)
		 (let* ((history-element (car history-elements))
			(guess (guess history-element))
			(ball-count (ball-count history-element)))
		   (recurse (cdr history-elements)
			    (cons guess result-guesses)
			    (cons (first ball-count) strikes)
			    (cons (second ball-count) balls))))))
    (recurse (guess-history algorithm) () () ())))
||#
(defun get-printable-history (algorithm)
  (let ((result-guesses nil)
	(strikes        nil)
	(balls          nil))
    (flet ((print-each-element (guess ball-count)
			       (push guess result-guesses)
			       (push (first  ball-count) strikes)
			       (push (second ball-count) balls)))
      (guess-history-iterate-element (guess-history algorithm) #'print-each-element))
    (values result-guesses strikes balls)))

(defun show-previous-guesses (algorithm)
  "Print a game over message with previous tries in reverse order."
  (multiple-value-bind (guesses strikes balls)
      (get-printable-history algorithm)
    ;;; FIXME: someday do FORMAT hacking!
    (format *query-io* "~&I solved it in ~D try, and my guesses were:" (length guesses))
    (map nil #'(lambda (guess strike ball)
		 (format *query-io*
			 "~&~10T~{~D~} ~@[: ~D Strikes ~D Balls~]~&"
			 guess
			 strike
			 ball))
	 guesses
	 strikes
	 balls)))

(defun show-guess-to-user (guess)
  "Print GUESS to the standard output"
  (format *query-io* "~{~D~}~&" guess))

(defun get-ball-count-str-from-user ()
  "Get input from user and return it as string"
  (read-line *query-io* nil nil))

(defun make-ball-count-from-string (str)
  "Return number of strikes and number of balls"
  (if (string-equal str "out")
      (values 0 0)
      (let ((strike-position (position #\s str :test #'char-equal))
	    (ball-position (position #\b str :test #'char-equal)))
	(cond ((and strike-position ball-position)
	       (values (parse-integer (subseq str 0 strike-position))
		       (parse-integer (subseq str (1+ strike-position) ball-position))))
	      (strike-position
	       (values (parse-integer (subseq str 0 strike-position))
		       0))
	      (ball-position
	       (values 0
		       (parse-integer (subseq str 0 ball-position))))))))


(defun show-unexpected-end-message (algorithm)
  "Print a game over message with previous tries in reverse order."
  (multiple-value-bind (guesses strikes balls)
      (get-printable-history algorithm)
    ;;; FIXME: someday do FORMAT hacking!
    (format *query-io* "~&I Can't guess the number in your mind, please check if you made any mistake:")
    (map nil #'(lambda (guess strike ball)
		 (format *query-io*
			 "~&~10T~{~D~} ~@[: ~D Strikes ~D Balls~]~&"
			 guess
			 strike
			 ball))
	 guesses
	 strikes
	 balls)))

;;; WARNING - try this on standalone listener.
;;; Sometimes I/O doesn't do right thing on Emacs lisp listener.

    
;;; Recursive version
(defun test-main1 (n)
  (let ((algorithm (make-guessing-algorithm 'simple-duplicate-permutation-algorithm n)))
    (labels
	((test-main-aux (guess)
			(show-guess-to-user guess)
			(multiple-value-bind (strikes balls)
			    (make-ball-count-from-string (get-ball-count-str-from-user))
			  (cond ((and strikes (= n strikes))
				 (show-previous-guesses algorithm))
				(t (improve-guess algorithm strikes balls)
				   (bind-if (guess (make-random-guess algorithm))
					    (test-main-aux guess)
					    (show-unexpected-end-message algorithm)))))))
      (bind-if (guess (make-random-guess algorithm))
	       (test-main-aux guess)
	       (show-unexpected-end-message algorithm)))))

;;; Loop version
(defun test-main2 (n)
  (let ((algorithm (make-guessing-algorithm 'simple-duplicate-permutation-algorithm n)))
    (loop as (strikes balls) = (multiple-value-list (make-ball-count-from-string
						   (get-ball-count-str-from-user)))
	until (and strikes (= strikes n))
	initially (bind-if (guess (make-random-guess algorithm))
			   (show-guess-to-user guess)
			   (progn (show-unexpected-end-message algorithm) (return)))
	do (improve-guess algorithm strikes balls)
	   (bind-if (guess (make-random-guess algorithm))
		    (show-guess-to-user guess)
		    (show-unexpected-end-message algorithm))		    
	finally (show-previous-guesses algorithm))))
